from __future__ import annotations

from enum import Enum, auto
import numpy as np
import xarray as xr
from dataclasses import dataclass
from cftime import DatetimeNoLeap
from sklearn.neighbors import BallTree



############################### TIME FUNCTIONS ################################

class Frequency(Enum):
    daily = auto()
    weekly = auto()
    monthly = auto()
    yearly = auto()
    decadal = auto()


def datetimeNoLeap_to_epoch(dt) -> float:
    """
    Convert a cftime.DatetimeNoLeap instance to an epoch timestamp.
    """
    # Define the reference datetime from which the epoch timestamp is counted.
    epoch = DatetimeNoLeap(1970, 1, 1)
    
    # Calculate the difference in seconds between the provided date and the epoch
    delta = dt - epoch
    
    # Convert the timedelta to seconds
    seconds = delta.days * 24 * 3600 + delta.seconds + delta.microseconds / 1e6

    return seconds


############################### SPACE/GEO FUNCTIONS ################################



class LongitudeConvention(Enum):
    ambiguous = auto()
    neg180_180 = auto()
    pos0_360 = auto()
    #TODO: consider adding -pi_+pi and 0_2pi
    # negπ_π = auto()
    # pos0_2π = auto()


#TODO: consider having a latitude convention
# class LatitudeConvention(Enum):
#     degrees = auto()
#     radians = auto()


@dataclass
class Resolution:
    dx: float
    dy: float = None

    def __init__(self, dx: float, dy: float|None=None):
        self.dx = dx
        self.dy = dy if dy is not None else dx


def validate_longitudes(lons:np.ndarray):
    """Validate that the given longitude values are in the range [-180, 360] and are monotonic (either positive or negative)"""
    assert np.all(lons >= -180) and np.all(lons <= 360), f'Longitude values must be in the range [-180, 360]. Got: {lons}'
    deltas = np.diff(lons)
    assert np.all(deltas >= 0) or np.all(deltas <= 0), f'Longitude values must be monotonic (either positive or negative). Got: {lons}'


def validate_latitudes(lats:np.ndarray):
    """Validate that the given latitude values are in the range [-90, 90], and are monotonic (either positive or negative)"""
    assert np.all(lats >= -90) and np.all(lats <= 90), f'Latitude values must be in the range [-90, 90]. Got: {lats}'
    deltas = np.diff(lats)
    assert np.all(deltas >= 0) or np.all(deltas <= 0), f'Latitude values must be monotonic (either positive or negative). Got: {lats}'

def determine_longitude_convention(lons:np.ndarray) -> LongitudeConvention:
    """Determine the longitude convention of the given longitude values"""

    # ensure valid longitude values
    validate_longitudes(lons)
    
    # determine the longitude convention
    if np.all(lons >= 0) and np.all(lons <= 180):
        return LongitudeConvention.ambiguous
    elif np.all(lons >= -180) and np.all(lons <= 180):
        return LongitudeConvention.neg180_180
    elif np.all(lons >= 0) and np.all(lons <= 360):
        return LongitudeConvention.pos0_360
    
    raise ValueError(f'Internal Error: Should be unreachable. Got: {lons}')
    
def convert_longitude_convention(lons:np.ndarray, target_convention:LongitudeConvention) -> np.ndarray:
    """Convert the given longitude values to the specified longitude convention"""

    assert np.all(lons >= -180) and np.all(lons <= 360), f'Longitude values must be in the range [-180, 360]. Got: {lons}'
    
    if target_convention == LongitudeConvention.ambiguous:
        target_convention = LongitudeConvention.neg180_180

    if target_convention == LongitudeConvention.neg180_180:
        return np.where(lons > 180, lons - 360, lons)
    elif target_convention == LongitudeConvention.pos0_360:
        return np.where(lons < 0, lons + 360, lons)
    else:
        raise ValueError(f'Invalid target longitude convention: {target_convention}. Expected one of: {[*LongitudeConvention.__members__.values()]}')


def inplace_set_longitude_convention(data:xr.DataArray, convention:LongitudeConvention) -> None:
    """
    Set the longitude convention of the data, and shift so that coordinates are monotonically increasing
    """
    if determine_longitude_convention(data['lon'].data) in [convention, LongitudeConvention.ambiguous]:
        return

    # map the longitude values to the new convention
    old_lon = data['lon'].data
    new_lon = convert_longitude_convention(old_lon, convention)

    # make lon the last axis for easier manipulation
    lon_idx = data.dims.index('lon')
    new_data = np.moveaxis(data.data, lon_idx, -1)

    # shift the data so that the lon coordinates are monotonically increasing
    sorted_indices = np.argsort(new_lon)
    new_lon = new_lon[sorted_indices]
    new_data = new_data[..., sorted_indices]

    # move the lon axis back, and make the data contiguous
    new_data = np.moveaxis(new_data, -1, lon_idx)
    new_data = np.ascontiguousarray(new_data)

    # update the data in place
    data.data = new_data
    data['lon'] = new_lon

def points_to_mask(lats: np.ndarray, lons: np.ndarray, /, n_lat=180, n_lon=360, min_lat=-90, max_lat=90, min_lon=-180, max_lon=180) -> xr.DataArray:
    #TODO: consider allowing for time to also be a coordinate
    """
    convert a list of points to a boolean mask that is true at each of the points

    Args:
    - lats (np.ndarray): an array of latitudes (in degrees)
    - lons (np.ndarray): an array of longitudes (in degrees)
    - n_lat (int, optional): the number of latitude bins for the output mask. Defaults to 180.
    - n_lon (int, optional): the number of longitude bins for the output mask. Defaults to 360.
    - min_lat (float, optional): the minimum latitude for the output mask. Defaults to -90.
    - max_lat (float, optional): the maximum latitude for the output mask. Defaults to 90.
    - min_lon (float, optional): the minimum longitude for the output mask. Defaults to -180.
    - max_lon (float, optional): the maximum longitude for the output mask. Defaults to 180.
    """

    # create grid
    grid_lats = np.linspace(min_lat, max_lat, n_lat+1)[:-1]
    grid_lons = np.linspace(min_lon, max_lon, n_lon+1)[:-1]

    # find the closest grid point to each point
    lat_diffs = np.abs(grid_lats - lats[:, None])
    lon_diffs = np.abs(grid_lons - lons[:, None])
    lat_idx = np.argmin(lat_diffs, axis=1)
    lon_idx = np.argmin(lon_diffs, axis=1)

    # create mask
    mask = np.zeros((n_lat, n_lon), dtype=bool)
    mask[lat_idx, lon_idx] = True

    data = xr.DataArray(mask, dims=['lat', 'lon'], coords={'lat': grid_lats, 'lon': grid_lons})

    return data

def mask_to_sdf(mask: xr.DataArray, include_initial_points:bool) -> xr.DataArray:
    """
    Generate a distance field from points in a mask

    Args:
    - mask (xr.DataArray): A boolean mask of the points to generate a distance field from. If float, non-zero values are considered True. NaNs are considered False.
    - include_initial_points (bool): Whether to include the initial points in the distance field. If False, the distance at the initial points will be NaN.
    """

    assert 'lat' in mask.coords and 'lon' in mask.coords, f'mask must have lat and lon coordinates. Got: {mask.coords.keys()}'

    # if mask only has lat/lon, it is already a single slice
    if mask.coords.keys() == {'lat', 'lon'}:
        sdf = mask_slice_to_sdf(mask.data, mask.lat, mask.lon, include_initial_points)
        return xr.DataArray(sdf, dims=['lat', 'lon'], coords={'lat': mask.lat, 'lon': mask.lon})


    #reshape mask data so that lat,lon are the last two dimensions
    mask_data = mask.data
    lat_dim_idx = mask.dims.index('lat')
    lon_dim_idx = mask.dims.index('lon')
    lon_lat_dim_idxs = [lat_dim_idx, lon_dim_idx]
    new_order = [i for i in range(len(mask.dims)) if i not in lon_lat_dim_idxs] + lon_lat_dim_idxs
    new_coords = [mask.dims[i] for i in new_order]
    mask_data = np.transpose(mask_data, axes=new_order)

    # create an empty array to store the distance field
    sdf_slices = np.zeros(mask_data.shape, dtype=float)

    # iterate over each slice of the mask
    lat, lon = mask.lat, mask.lon
    for idx in np.ndindex(mask_data.shape[:-2]):
        slice_data = mask_data[idx]
        sdf_slices[idx] = mask_slice_to_sdf(slice_data, lat, lon, include_initial_points)

    # create the new DataArray
    sdf = xr.DataArray(sdf_slices, dims=new_coords, coords={dim: mask.coords[dim] for dim in new_coords})
    return sdf



def mask_slice_to_sdf(mask: np.ndarray, lat:np.ndarray, lon:np.ndarray, include_initial_points:bool) -> np.ndarray:
    """
    Generate a distance field from points in a single (lat/lon) mask slice

    Args:
    - mask (np.ndarray): A boolean mask of the points to generate a distance field from. If float, non-zero values are considered True. NaNs are considered False.
    """

    # collect just the True the points from the mask (and convert to radians)
    assert len(mask.shape) == 2, f'mask must be 2D. Got: {mask.shape}'
    assert len(lat) == mask.shape[0] and len(lon) == mask.shape[1], f'lat and lon must have the same length as the mask. Got: {len(lat)}, {len(lon)} and {mask.shape}'

    mask_data = np.nan_to_num(mask, nan=0)
    points_x_idx, points_y_idx = np.argwhere(mask_data).T
    points_x = lat[points_x_idx]
    points_y = lon[points_y_idx]
    points = np.stack([points_x, points_y], axis=1)
    points = np.deg2rad(points)

    # create the list of points to compute the distance field for (literally just mask's coordinates)
    mesh = np.stack(np.meshgrid(lat, lon), axis=2)
    mesh_shape = mesh.shape[:2]
    mesh = np.deg2rad(mesh)
    mesh = mesh.reshape(-1, 2)

    # efficiently compute the closest point in the mask to each point in the mesh
    tree = BallTree(points, metric='haversine')
    sdf = tree.query(mesh)[0]
    sdf = sdf.reshape(*mesh_shape).T  # reshape and put lat as first dimension
    sdf *= 6371.0  # convert from radians to kilometers

    # set the distance at the initial points to NaN
    if not include_initial_points:
        sdf[points_x_idx, points_y_idx] = np.nan

    # if original mask contains NaNs, preserve them in the distance field
    if np.any(np.isnan(mask)):
        sdf[np.isnan(mask)] = np.nan

    return sdf




import os
import requests
from pathlib import Path
from zipfile import ZipFile
from tqdm import tqdm
import geofeather as gf
from geopandas import GeoDataFrame


from functools import lru_cache
cache = lru_cache(maxsize=None)


import pdb


default_gadm_path = Path(__file__).parent/'gadm'
gadm_path = None
gadm_env_var = 'GADM_DIR'


GADM_FILES = [
    'gadm36_0.feather',
    'gadm36_2.feather',
    'gadm36_3.feather'
]

def verify_gadm_dir(dir:Path):
    """Presuming that GADM data is already downloaded, verify that the given directory contains the GADM data in the expected format"""
    #verify that the directory exists
    assert dir.is_dir(), f'Invalid GADM directory. Expected a directory. Got: {dir}'

    for filename in GADM_FILES:
        verify_gadm_file(dir/filename)


def verify_gadm_file(file:Path):
    """Verify that the given file is a valid GADM file"""
    level = file.stem.split('_')[-1]
    assert file.is_file(), f'Invalid GADM directory. No admin{level} data ({file.name}) found in {file.parent}'
    gf.from_geofeather(file)


def download_gadm(dir:Path):
    """Download GADM data to the given directory"""

    # create the directory if it doesn't exist
    os.makedirs(dir, exist_ok=True)

    # download each file
    root = 'https://jataware-world-modelers.s3.amazonaws.com/gadm/'
    for filename in GADM_FILES:
        url = f'{root}{filename}.zip'
        download_gadm_file(url, filename, dir)


def download_gadm_file(url:str, filename:str, dir:Path):
    """Download a GADM file from the S3 bucket"""
    # check if the file already exists
    if (dir/filename).is_file():
        print(f'{filename} already exists. Skipping download')
        return

    # add the .zip extension to the filename
    assert filename.endswith('.feather'), f'Invalid filename. Expected a .feather file. Got: {filename}'
    filename = f'{filename}.zip'

    # download the file
    r = requests.get(url, allow_redirects=True, stream=True)
    assert r.status_code == 200, f'Failed to download GADM data from {url}. Got status code: {r.status_code}'

    # get the total size of the file for the progress bar
    total_size = int(r.headers.get('content-length', 0))

    # save the file
    with open(dir/filename, 'wb') as f, tqdm(desc=f'download {filename}', total=total_size, unit='iB', unit_scale=True) as bar:
        for chunk in r.iter_content(chunk_size=1024):
            f.write(chunk)
            bar.update(len(chunk))

    # unzip the file
    unzip_gadm_file(filename, dir)


def unzip_gadm_file(filename:str, dir:Path):
    """Unzip a GADM file with a progress bar into the given directory"""

    try:
        with ZipFile(dir/filename, 'r') as z:
            # Get the total uncompressed size for the progress bar
            total_size = sum(item.file_size for item in z.infolist())
            with tqdm(desc=f'unzip {filename}', total=total_size, unit='B', unit_scale=True, unit_divisor=1024) as bar:
                
                for member in z.infolist():
                    # Extract each member in chunks to update the progress bar
                    with z.open(member, 'r') as source, open(dir/member.filename, 'wb') as target:
                        chunk_size = 1024 * 1024  # 1MB
                        while True:
                            chunk = source.read(chunk_size)
                            if not chunk:
                                break
                            target.write(chunk)
                            bar.update(len(chunk))

    # delete the zip file
    finally:
        os.remove(dir/filename)

def setup_gadm():
    """Ensure that GADM data is downloaded and ready to use"""
    global gadm_path
    # if GADM_DIR environment variable is set, use that
    _gadm_path = os.environ.get(gadm_env_var)
    if _gadm_path is not None:
        gadm_path = Path(_gadm_path)
        verify_gadm_dir(gadm_path)
        print(f'Using ENV specified GADM data from {gadm_path}')
        return

    # otherwise use the default directory, and download GADM if necessary
    gadm_path = default_gadm_path
    try:
        verify_gadm_dir(gadm_path)
        print(f'Using default GADM data from {gadm_path}')
        return
    except:
        print(f'Downloading default GADM data to {gadm_path}')
        download_gadm(gadm_path)
        verify_gadm_dir(gadm_path)
        print(f'Using default GADM data from {gadm_path}')

@cache
def get_admin0_shapes():
    assert gadm_path is not None, f'GADM data not initialized. Call setup_gadm() first'
    shapes = gf.from_geofeather(gadm_path/'gadm36_0.feather')
    return shapes

# @cache
# def get_admin1_shapes():
#     raise NotImplementedError("TODO: aggregate admin1 shapes from admin2 shapes")

@cache
def get_admin2_shapes() -> GeoDataFrame:
    assert gadm_path is not None, f'GADM data not initialized. Call setup_gadm() first'
    shapes = gf.from_geofeather(gadm_path/'gadm36_2.feather')
    return shapes

@cache
def get_admin3_shapes() -> GeoDataFrame:
    assert gadm_path is not None, f'GADM data not initialized. Call setup_gadm() first'
    shapes = gf.from_geofeather(gadm_path/'gadm36_3.feather')
    return shapes

# _sf = None
# def get_gadm():
#     """Get the country shapefile"""
#     if _sf is None:
#         self.print(f'Loading country shapefile...')
#         _sf = gpd.read_file(f'{dirname(abspath(__file__))}/gadm_0/gadm36_0.shp')
#     return self._sf


if __name__ == '__main__':
    #DEBUG
    # download_gadm(default_gadm_path)
    # get_admin2_shapes()
    get_admin3_shapes()


from __future__ import annotations

from matplotlib import pyplot as plt
import numpy as np
import xarray as xr
from rasterio.transform import from_bounds
from rasterio.features import geometry_mask
import calendar

from .spacetime import Frequency, Resolution, DatetimeNoLeap, LongitudeConvention, inplace_set_longitude_convention, mask_to_sdf
from .regrid import RegridType, regrid_1d
from .utilities import method_uses_prop, xarray_mode
from .gadm import setup_gadm, get_admin2_shapes, get_admin3_shapes

from itertools import count
from inspect import signature, Signature
from enum import Enum, auto
from typing import Any, Callable, TypeVar, Mapping, Literal
from typing_extensions import ParamSpec
from types import MethodType
from dataclasses import dataclass


# pretty printing
from rich import print, traceback; traceback.install(show_locals=True)


import pdb




class ThresholdType(Enum):
    greater_than = auto()
    less_than = auto()
    greater_than_or_equal = auto()
    less_than_or_equal = auto()
    equal = auto()
    not_equal = auto()

@dataclass
class Threshold:
    value: float|int
    type: ThresholdType
    is_percentile: bool = False


# Datatypes for representing identifiers in the pipeline namespace
class ResultID(str): ...
class OperandID(str): ...


@dataclass
class Variable:
    data: xr.DataArray
    time_regrid_type: RegridType|None
    geo_regrid_type: RegridType|None

@dataclass
class PipelineVariable(Variable):
    frequency: Frequency|str
    resolution: Resolution|str
    
    @staticmethod
    def from_result(
        data:xr.DataArray,
        prev:'PipelineVariable',
        *,
        time_regrid_type:RegridType=None,
        geo_regrid_type:RegridType=None,
        frequency:Frequency|str=None,
        resolution:Resolution|str=None
    ) -> 'PipelineVariable':
        """Create a new Variable from the given data, inheriting the geo/temporal resolution from a previous Variable"""
        return PipelineVariable(
            data,
            time_regrid_type or prev.time_regrid_type,
            geo_regrid_type or prev.geo_regrid_type,
            frequency or prev.frequency,
            resolution or prev.resolution
        )
    
    @classmethod
    def from_base_variable(cls, var:Variable, frequency:Frequency|str, resolution:Resolution|str) -> 'PipelineVariable':
        return cls(var.data, var.time_regrid_type, var.geo_regrid_type, frequency, resolution)
    

_R_co = TypeVar("_R_co", covariant=True)
_P = ParamSpec("_P")

# More correct type for @compile decorator (correctly handling .unwrapped property), but doesn't show the method description in the IDE
# class CallableWithUnwrapped(Protocol[_P, _R_co], Callable[_P, _R_co]):
#     # def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R_co: ...
    
#     @property
#     def unwrapped(self) -> Callable[_P, _R_co]: ...

class Pipeline:


    def __init__(self, *, verbose:bool=True, low_memory:bool=False):
        
        # dynamic settings for data to be used in pipeline
        self.resolution: Resolution|str|None = None
        self.frequency: Frequency|str|None = None
        
        # list of steps in the pipeline DAG
        self.steps: list[tuple[MethodType, tuple[Any, ...]]] = []

        # keep track of the most recent result during pipeline execution
        self.last_set_identifier: str|None = None

        # namespace for binding operation results to identifiers 
        self.env: dict[str, PipelineVariable] = {}

        # keep track of all identifiers while the pipeline is being compiled
        self.compiled_ids: set[str] = set()

        # tmp id counter
        self.tmp_id_counter = count(0)

        # whether to print out debug info during pipeline execution
        self.verbose = verbose

        # whether to use low memory mode
        self.low_memory = low_memory

        # bind the current instance to all unwrapped compiled methods (so we don't need to pass the instance manually)
        for attr_name, attr_value in vars(Pipeline).items():
            if hasattr(attr_value, "unwrapped"):
                bound_unwrapped = attr_value.unwrapped.__get__(self, Pipeline)
                setattr(attr_value, "unwrapped", bound_unwrapped)

        # shapefile for country data. Only load if needed
        self._sf = None


    def compile(method:Callable[_P, _R_co]) -> Callable[_P, None]:
        """
        Decorator to make a method a compile-time method.

        Adds the method to the list of steps in the pipeline (instead of running it)
        checks the the input identifiers already exist in the pipeline namespace
        checks that the result identifier is unique and marks it as used

        NOTE: All identifiers in compiled functions must use either the `ResultID` or `OperandID` type
        identifiers that are not correctly marked will not be properly checked during compilation
        Also all identifiers must be positional arguments, not keyword arguments
        """

        # determine the function signature index of the result and operand identifiers if any
        sig = signature(method)
        params = list(sig.parameters.values())
        result_idx = []
        operand_idx = []
        for i, param in enumerate(params[1:]): #skip self
            if param.annotation == ResultID:
                assert param.kind == param.POSITIONAL_ONLY, f'ERROR compiling "Pipeline.{method.__name__}". Identifier "{param.name}:{param.annotation}" must be a positional-only argument'
                result_idx.append(i)
            elif param.annotation == OperandID:
                assert param.kind == param.POSITIONAL_ONLY, f'ERROR compiling "Pipeline.{method.__name__}". Identifier "{param.name}:{param.annotation}" must be a positional-only argument'
                operand_idx.append(i)


        #TODO: perhaps we don't need this restriction
        assert len(result_idx) <= 1, f'ERROR compiling "Pipeline.{method.__name__}". Compiled functions can only have one result identifier'


        # Check the return annotation
        if sig.return_annotation not in {None, Signature.empty}:
            raise ValueError(f'Method "{method.__name__}" should not have a return annotation or it should be set to None.')

        def wrapper(self:'Pipeline', *args, **kwargs):

            # append the function to the list of steps in the pipeline
            self.steps.append((method, (self,) + args, kwargs))

            # check that any operand identifiers exist in the pipeline namespace
            for i in operand_idx:
                self._assert_id_exists(args[i])
            
            # check that the result identifier is unique and mark it as used
            for i in result_idx:
                self._assert_id_is_unique_and_mark_used(args[i])

            # check if the method uses any adminN properties, meaning we need to verify gadm shapefiles are available
            if any(method_uses_prop(Pipeline, method, name, key=lambda m: m.unwrapped) for name in ['admin0', 'admin1', 'admin2', 'admin3']):
                setup_gadm()

        # save the unmodified original function in for use inside pipeline methods
        wrapper.unwrapped = method

        return wrapper

    
    @staticmethod
    def step_str(step:tuple[MethodType, tuple[Any, ...], dict[str, Any]]):
        """get the repr for the given step in the pipeline"""
        func, args, kwargs = step
        args = args[1:] #skip self. Removing this causes infinite recursion
        name = func.__name__
        args_str = ', '.join([str(arg) for arg in args])
        kwargs_str = ', '.join([f'{key}={value}' for key, value in kwargs.items()])
        return f'Pipeline.{name}({", ".join((args_str, kwargs_str))})'
    

    def __str__(self):
        """get the str for the pipeline"""
        return '\n'.join([f'{i}: {Pipeline.step_str(step)}' for i, step in enumerate(self.steps)])

    def __repr__(self):
        """get the repr for the pipeline"""
        return f'Pipeline({[func.__name__ for func, _, _ in self.steps]})'

    
    def _next_tmp_id(self) -> str:
        """(runtime) get the next available tmp id (to be used for intermediate results)"""
        while True: 
            tmp_id = f'__tmp_{next(self.tmp_id_counter)}__'
            if tmp_id not in self.compiled_ids:
                return tmp_id

    
    def _assert_id_exists(self, identifier:str):
        """
        (Compile-time) assert that the given identifier exists in the pipeline namespace
        Should be called inside any compile-time functions that will use a variable from the pipeline namespace
        NOTE: always call after appending the current step to the pipeline

        @compile handles this automatically
        """
        if identifier not in self.compiled_ids:
            pdb.set_trace()
            raise ValueError(f'Operand identifier "{identifier}" does not exist in pipeline at step {len(self.steps)-1}: {self.step_str(self.steps[-1])}')
    
    def _assert_id_is_unique_and_mark_used(self, identifier:str):
        """
        (Compile-time) assert that the given identifier is unique, and add it to the compiled_ids set
        Should be called inside any compile-time functions that will add a variable to the pipeline namespace
        NOTE: always call after appending the current step to the pipeline

        @compile handles this automatically
        """
        if identifier in self.compiled_ids:
            raise ValueError(f'Tried to reuse "{identifier}" for result identifier on step {len(self.steps)-1}: {self.step_str(self.steps[-1])}. All identifiers must be unique.')
        self.compiled_ids.add(identifier)

    
    
    def bind_value(self, identifier:ResultID, value:PipelineVariable):
        """Bind a value to an identifier in the pipeline namespace"""
        assert identifier not in self.env, f'Identifier "{identifier}" already exists in pipeline namespace. All identifiers must be unique.'
        self.last_set_identifier = identifier
        self.env[identifier] = value


    def get_value(self, identifier:OperandID) -> PipelineVariable:
        """Get a value from the pipeline namespace"""
        return self.env[identifier]
    

    def get_last_value(self) -> PipelineVariable:
        """Get the last value that was set in the pipeline namespace"""
        assert self.last_set_identifier is not None, 'No value has been set yet'
        return self.env[self.last_set_identifier]


    @compile
    def set_geo_resolution(self, target:Resolution|str):
        """
        Set the current target geo resolution for the pipeline

        Resolution can either be a fixed Resolution object, e.g. Resolution(0.5, 0.5), 
        or target the resolution of an existing dataset in the pipeline by name, e.g. 'tasmax'

        NOTE: resolution updates are only applied to results of operations, not on data load.
        """
        self.resolution = target


    @compile
    def set_time_resolution(self, target:Frequency|str):
        """
        Set the current target temporal resolution for the pipeline

        Frequency can either be a fixed Frequency object, e.g. Frequency.monthly,
        or target the frequency of an existing dataset in the pipeline by name, e.g. 'tasmax'

        NOTE: frequency updates are only applied to results of operations, not on data load.
        """
        self.frequency = target

    
    def _assert_pipe_has_resolution(self):
        """
        check if the pipeline has already set a target geo and temporal resolution
        Should be called before adding any binary operations to the pipeline
        """
        time_res, geo_res = False, False
        for func, _ in self.steps:
            if func == self.set_time_resolution:
                time_res = True
            elif func == self.set_geo_resolution:
                geo_res = True
        if not time_res:
            raise ValueError('Pipeline must have a target temporal resolution before performing binary operations')
        if not geo_res:
            raise ValueError('Pipeline must have a target geo resolution before performing binary operations')

    
    @compile
    def load(self, name:ResultID, /, loader:Callable[[], Variable]):
        """
        Load existing data into the pipeline

        Args:
            identifier (str): the identifier to bind the data to in the pipeline namespace
            data (enum): the data to load
            model (enum, optional): the model to load the data from (required for CMIP6 data)
        """
        var = loader()

        #ensure the lon conventions match
        if 'lon' in var.data.coords:
            inplace_set_longitude_convention(var.data, LongitudeConvention.neg180_180)
        #TODO: match the latitude convention

        self.bind_value(name, PipelineVariable.from_base_variable(var, name, name))


    @compile
    def threshold(self, y:ResultID, x:OperandID, /, threshold:Threshold, preserve_nan:bool=True):
        """
        Threshold a dataset, e.g. `y = x > threshold` or `y = x <= threshold`

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to threshold
            threshold (Threshold): the threshold to apply to the data
            preserve_nan (bool, optional): whether to output the result as a float array (with NaNs preserved) or bool array.
                if True, then result will be cast as a float array, and any NaNs from the original will be propagated.
                if False, then original bool array of threshold operation will be returned, and NaNs will be set to False.
                Defaults to True.
        """

        #first make sure the data matches the specified resolution and frequency
        x = self.auto_regrid(x, allow_no_target=True)
        
        # grab the data
        var = self.get_value(x)

        # if percentile threshold, calculate the threshold value
        if threshold.is_percentile:
            #nanpercentile doesn't handle inf, so do it manually
            non_nan_data = var.data.data[~(np.isnan(var.data.data) | np.isinf(var.data.data))]
            threshold_value = np.percentile(non_nan_data, threshold.value)
        else:
            threshold_value = threshold.value
        
        # perform the threshold operation
        if threshold.type == ThresholdType.greater_than:
            result = var.data > threshold_value
        elif threshold.type == ThresholdType.less_than:
            result = var.data < threshold_value
        elif threshold.type == ThresholdType.greater_than_or_equal:
            result = var.data >= threshold_value
        elif threshold.type == ThresholdType.less_than_or_equal:
            result = var.data <= threshold_value
        elif threshold.type == ThresholdType.equal:
            result = var.data == threshold_value
        elif threshold.type == ThresholdType.not_equal:
            result = var.data != threshold_value
        else:
            raise ValueError(f'Unrecognized threshold type: {threshold.type}. Expected one of: {[*ThresholdType.__members__.values()]}')

        if preserve_nan:
            result = result.astype(float)
            result = result.where(~np.isnan(var.data))

        # save the result to the pipeline namespace
        self.bind_value(y, PipelineVariable.from_result(result, var))



    @compile
    def fixed_time_regrid(self, y:ResultID, x:OperandID, /, target:Frequency):
        """
        regrid the given data to the given fixed temporal frequency
        
        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to regrid
            target (Frequency): the fixed target temporal frequency to regrid to, e.g. Frequency.monthly
        """
        var = self.get_value(x)
        if var.frequency == target or var.time_regrid_type is None or 'time' not in var.data.dims:
            #skip regridding
            self.bind_value(y, var)
            return
        
        # generate the new time coordinates
        min_time = var.data.time.data.min()
        max_time = var.data.time.data.max()
        if target == Frequency.daily:
            min_time = min_time.replace(hour=0, minute=0, second=0, microsecond=0)
            max_time = max_time.replace(hour=0, minute=0, second=0, microsecond=0)
            times = np.array([DatetimeNoLeap(year, month, day) for year in range(min_time.year, max_time.year+1) for month in range(1, 13) for day in range(1, calendar.monthrange(year, month)[1]+1)])
        elif target == Frequency.weekly:
            min_time = min_time.replace(hour=0, minute=0, second=0, microsecond=0)
            max_time = max_time.replace(hour=0, minute=0, second=0, microsecond=0)
            times = np.array([DatetimeNoLeap(year, month, day) for year in range(min_time.year, max_time.year+1) for month in range(1, 13) for day in range(1, calendar.monthrange(year, month)[1]+1) if calendar.weekday(year, month, day) == calendar.SUNDAY])
        elif target == Frequency.monthly:
            min_time = min_time.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
            max_time = max_time.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
            times = np.array([DatetimeNoLeap(year, month, 1) for year in range(min_time.year, max_time.year+1) for month in range(1, 13)])
        elif target == Frequency.yearly:
            min_time = min_time.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
            max_time = max_time.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
            times = np.array([DatetimeNoLeap(year, 1, 1) for year in range(min_time.year, max_time.year+1)])
        elif target == Frequency.decadal:
            min_time = DatetimeNoLeap(min_time.year - min_time.year % 10, 1, 1)
            max_time = DatetimeNoLeap(max_time.year + max_time.year % 10, 1, 1)
            times = np.array([DatetimeNoLeap(year, 1, 1) for year in range(min_time.year, max_time.year+1, 10)])
        else:
            raise ValueError(f'Invalid target temporal frequency: {target}. Expected one of: {[*Frequency.__members__.values()]}')

        # regrid the data and save the result to the pipeline namespace
        new_data = regrid_1d(var.data, times, 'time', aggregation=var.time_regrid_type, low_memory=self.low_memory)
        var = PipelineVariable.from_result(new_data, var, frequency=target)
        self.bind_value(y, var)

        
        
    @compile
    def matched_time_regrid(self, y:ResultID, x:OperandID, /, target:str):
        """
        regrid the given data to match the temporal frequency of the specified target data
        
        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to regrid
            target (str): the identifier for the target data to match temporal frequency with while regridding
                should target the resolution of an existing dataset in the pipeline by name, e.g. 'tasmax'
        """
        var = self.get_value(x)
        target_var = self.get_value(target)

        if var.frequency == target_var.frequency or var.time_regrid_type is None or 'time' not in var.data.dims:
            #skip regridding for already matching data, or data has no time dimension
            self.bind_value(y, var)
            return

        new_data = regrid_1d(var.data, target_var.data.time.data, 'time', aggregation=var.time_regrid_type, low_memory=self.low_memory)
        var = PipelineVariable.from_result(new_data, var, frequency=target_var.frequency)
        self.bind_value(y, var)


    @compile
    def fixed_geo_regrid(self, y:ResultID, x:OperandID, /, target:Resolution):
        """
        regrid the given data to the given fixed geo resolution

        Args:
            result (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to regrid
            target (Resolution): the fixed target geo resolution to regrid to, e.g. Resolution(0.5, 0.5),
        """
        var = self.get_value(x)
        if var.resolution == target or var.geo_regrid_type is None or 'lat' not in var.data.dims or 'lon' not in var.data.dims:
            #skip regridding
            self.bind_value(y, var)
            return

        # raw geo coordinates at target resolution
        lats = np.arange(-90, 90, target.dy)
        lons = np.arange(-180, 180, target.dx)

        # crop geo coordinates around the data's maximum extents
        min_lat = var.data.lat.data.min()
        max_lat = var.data.lat.data.max()
        min_lon = var.data.lon.data.min()
        max_lon = var.data.lon.data.max()
        lats = lats[(lats + target.dy/2 >= min_lat) & (lats - target.dy/2 <= max_lat)]
        lons = lons[(lons + target.dx/2 >= min_lon) & (lons - target.dx/2 <= max_lon)]

        # regrid the data and save the result to the pipeline namespace
        new_data = regrid_1d(var.data, lats, 'lat', aggregation=var.geo_regrid_type, low_memory=self.low_memory)
        new_data = regrid_1d(new_data, lons, 'lon', aggregation=var.geo_regrid_type, low_memory=self.low_memory)
        var = PipelineVariable.from_result(new_data, var, resolution=target)
        self.bind_value(y, var)
    
    @compile
    def matched_geo_regrid(self, y:ResultID, x:OperandID, /, target:str):
        """
        regrid the given data to match the geo resolution of the specified target data

        Args:
            result (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to regrid
            target (str): the identifier for the target data to match geo resolution while regridding
                should target the resolution of an existing dataset in the pipeline by name, e.g. 'tasmax'
        """
        var = self.get_value(x)
        target_var = self.get_value(target)

        if var.resolution == target_var.resolution or var.geo_regrid_type is None or 'lat' not in var.data.dims or 'lon' not in var.data.dims:
            #skip regridding for already matching data, or data has no geo dimensions
            self.bind_value(y, var)
            return

        new_data = regrid_1d(var.data, target_var.data.lat.data, 'lat', aggregation=var.geo_regrid_type, low_memory=self.low_memory)
        new_data = regrid_1d(new_data, target_var.data.lon.data, 'lon', aggregation=var.geo_regrid_type, low_memory=self.low_memory)
        var = PipelineVariable.from_result(new_data, var, resolution=target_var.resolution)
        self.bind_value(y, var)


    def auto_regrid(self, x:OperandID, allow_no_target:bool=False) -> OperandID:
        """
        automatically regrid the given data to the target geo and temporal resolution of the pipeline

        Args:
            x (str): the identifier of the data to regrid
            allow_no_target (bool, optional): whether the pipeline may skip target geo/temporal resolutions. Defaults to False.

        Returns:
            str: the identifier of the regridded data
        """

        # check if geo/time targets are necessary
        if not allow_no_target:
            assert self.resolution is not None, f'Pipeline must have a target geo resolution before performing regrid operation on "{x}"'
            assert self.frequency is not None, f'Pipeline must have a target temporal resolution before performing regrid operation on "{x}"'

        #TODO: need to do regridding operations that reduce size of data before ones that increase it
        # perform the time regrid
        if self.frequency is not None and self.env[x].frequency != self.frequency and 'time' in self.env[x].data.dims:
            tmp_id = self._next_tmp_id()
            if isinstance(self.frequency, Frequency):
                self.print(f'AUTO-REGRIDDING "{x}" TIME TO {self.frequency}')
                self.fixed_time_regrid.unwrapped(tmp_id, x, self.frequency)
            else:
                self.print(f'AUTO-REGRIDDING "{x}" TO TIME MATCH "{self.frequency}"')
                self.matched_time_regrid.unwrapped(tmp_id, x, self.frequency)
            x = tmp_id
        
        # perform the geo regrid
        if self.resolution is not None and self.env[x].resolution != self.resolution and 'lat' in self.env[x].data.dims and 'lon' in self.env[x].data.dims:
            tmp_id = self._next_tmp_id()
            if isinstance(self.resolution, Resolution):
                self.print(f'AUTO-REGRIDDING "{x}" GEO TO {self.resolution}')
                self.fixed_geo_regrid.unwrapped(tmp_id, x, self.resolution)
            else:
                self.print(f'AUTO-REGRIDDING "{x}" TO GEO MATCH "{self.resolution}"')
                self.matched_geo_regrid.unwrapped(tmp_id, x, self.resolution)
            x = tmp_id

        return x
    

    def binop(self, y:ResultID, x1:OperandID, x2:OperandID, op:Callable[[xr.DataArray, xr.DataArray], xr.DataArray]):
        """
        Perform a binary operation on two datasets, e.g. `y = x1 + x2`

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x1 (str): the identifier of the left operand
            x2 (str): the identifier of the right operand
            op (Callable[[xr.DataArray, xr.DataArray], xr.DataArray]): the binary operation to perform
        """

        #ensure data matches the specified resolution and frequency
        x1 = self.auto_regrid(x1)
        x2 = self.auto_regrid(x2)

        # grab the data values
        var1 = self.get_value(x1)
        data1 = var1.data
        var2 = self.get_value(x2)
        data2 = var2.data

        # perform the operation
        result = op(data1, data2)

        # save the result to the pipeline namespace
        self.bind_value(y, PipelineVariable.from_result(result, var1))
    
    @compile
    def add(self, y:ResultID, x1:OperandID, x2:OperandID, /):
        """
        Add two datasets together, i.e. `y = x1 + x2`

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x1 (str): the identifier of the left operand
            x2 (str): the identifier of the right operand
        """
        self.binop(y, x1, x2, lambda x1, x2: x1 + x2)


    @compile
    def subtract(self, y:ResultID, x1:OperandID, x2:OperandID, /):
        """
        Subtract two datasets, i.e. `y = x1 - x2`

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x1 (str): the identifier of the left operand
            x2 (str): the identifier of the right operand
        """
        self.binop(y, x1, x2, lambda x1, x2: x1 - x2)


    @compile
    def multiply(self, y:ResultID, x1:OperandID, x2:OperandID, /):
        """
        Multiply two datasets together, i.e. `y = x1 * x2`

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x1 (str): the identifier of the left operand
            x2 (str): the identifier of the right operand
        """
        self.binop(y, x1, x2, lambda x1, x2: x1 * x2)


    @compile
    def divide(self, y:ResultID, x1:OperandID, x2:OperandID, /):
        """
        Divide two datasets, i.e. `y = x1 / x2`

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x1 (str): the identifier of the left operand
            x2 (str): the identifier of the right operand
        """
        self.binop(y, x1, x2, lambda x1, x2: x1 / x2)


    @compile
    def power(self, y:ResultID, x1:OperandID, x2:OperandID, /):
        """
        Raise the first dataset to the power of the second dataset, i.e. `y = x1 ** x2`

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x1 (str): the identifier of the left operand
            x2 (str): the identifier of the right operand
        """
        self.binop(y, x1, x2, lambda x1, x2: x1 ** x2)

    # def and(self, y:ResultID, x1:OperandID, x2:OperandID, /):
    # def or(self, y:ResultID, x1:OperandID, x2:OperandID, /):
    # def xor(self, y:ResultID, x1:OperandID, x2:OperandID, /):
    # def not(self, y:ResultID, x:OperandID, /):


    @compile
    def sum_reduce(self, y:ResultID, x:OperandID, /, dims:list[str]):
        """
        Sum the data along the given dimensions

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to sum
            dims (list[str]): the list of dimensions to sum over
        """
        var = self.get_value(x)
        result = var.data.sum(dim=dims)
        self.bind_value(y, PipelineVariable.from_result(result, var))


    @compile
    def mean_reduce(self, y:ResultID, x:OperandID, /, dims:list[str]):
        """
        Take the mean of the data along the given dimensions

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to take the mean of
            dims (list[str]): the list of dimensions to take the mean over
        """
        var = self.get_value(x)
        result = var.data.mean(dim=dims)
        self.bind_value(y, PipelineVariable.from_result(result, var))
    
    

    @compile
    def max_reduce(self, y:ResultID, x:OperandID, /, dims:list[str]):
        """
        Take the maximum value of the data along the given dimensions

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to take the maximum value of
            dims (list[str]): the list of dimensions to take the maximum value over
        """
        var = self.get_value(x)
        result = var.data.max(dim=dims)
        self.bind_value(y, PipelineVariable.from_result(result, var))

    
    @compile
    def min_reduce(self, y:ResultID, x:OperandID, /, dims:list[str]):
        """
        Take the minimum value of the data along the given dimensions

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to take the minimum value of
            dims (list[str]): the list of dimensions to take the minimum value over
        """
        var = self.get_value(x)
        result = var.data.min(dim=dims)
        self.bind_value(y, PipelineVariable.from_result(result, var))

    
    @compile
    def std_reduce(self, y:ResultID, x:OperandID, /, dims:list[str]):
        """
        Take the standard deviation of the data along the given dimensions

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to take the standard deviation of
            dims (list[str]): the list of dimensions to take the standard deviation over
        """
        var = self.get_value(x)
        result = var.data.std(dim=dims)
        self.bind_value(y, PipelineVariable.from_result(result, var))
    

    @compile
    def median_reduce(self, y:ResultID, x:OperandID, /, dims:list[str]):
        """
        Take the median value of the data along the given dimensions

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to take the median value of
            dims (list[str]): the list of dimensions to take the median value over
        """
        var = self.get_value(x)
        result = var.data.median(dim=dims)
        self.bind_value(y, PipelineVariable.from_result(result, var))
    

    @compile
    def mode_reduce(self, y:ResultID, x:OperandID, /, dims:list[str]):
        """
        Take the mode of the data along the given dimensions

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to take the mode of
            dims (list[str]): the list of dimensions to take the mode over
        """
        var = self.get_value(x)
        result = xarray_mode(var.data, dim=dims)
        self.bind_value(y, PipelineVariable.from_result(result, var))

    @compile
    def scalar_add(self, y:ResultID, x:OperandID, /, scalar:float|int):
        """
        Add a scalar to the given data

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to add the scalar to
            scalar (float|int): the scalar to add to the data
        """
        var = self.get_value(x)
        result = var.data + scalar
        self.bind_value(y, PipelineVariable.from_result(result, var))


    @compile
    def scalar_subtract(self, y:ResultID, x:OperandID, /, scalar:float|int):
        """
        Subtract a scalar from the given data

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to subtract the scalar from
            scalar (float|int): the scalar to subtract from the data
        """
        var = self.get_value(x)
        result = var.data - scalar
        self.bind_value(y, PipelineVariable.from_result(result, var))


    @compile
    def scalar_multiply(self, y:ResultID, x:OperandID, /, scalar:float|int):
        """
        Multiply the given data by a scalar

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to multiply by the scalar
            scalar (float|int): the scalar to multiply the data by
        """
        var = self.get_value(x)
        result = var.data * scalar
        self.bind_value(y, PipelineVariable.from_result(result, var))


    @compile
    def scalar_divide(self, y:ResultID, x:OperandID, /, scalar:float|int, position:Literal['numerator', 'denominator']):
        """
        Divide the given data by a scalar

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to divide by the scalar
            scalar (float|int): the scalar to divide the data by
            position (Literal['numerator', 'denominator']): the position of the scalar in the division operation
        """
        var = self.get_value(x)
        if position == 'numerator':
            result = scalar / var.data
        else:
            result = var.data / scalar
        self.bind_value(y, PipelineVariable.from_result(result, var))


    @compile
    def scalar_power(self, y:ResultID, x:OperandID, /, scalar:float|int, position:Literal['base', 'exponent']):
        """
        Raise the given data to the power of a scalar

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to raise to the power of the scalar
            scalar (float|int): the scalar to raise the data to the power of
            position (Literal['base', 'exponent']): the position of the scalar in the power operation
        """
        var = self.get_value(x)
        if position == 'base':
            result = scalar ** var.data
        else:
            result = var.data ** scalar
        self.bind_value(y, PipelineVariable.from_result(result, var))


    @compile
    def isel(self, y:ResultID, x:OperandID, /, indexers:Mapping[Any, Any]|None=None, drop:bool=False):
        """
        Perform an isel operation on the given data. 
        See: https://xarray.pydata.org/en/stable/generated/xarray.DataArray.isel.html

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to perform the isel operation on
            indexers (Mapping[Any, Any], optional): A dict with keys matching dimensions and values given by integers, slice objects or arrays. indexer can be a integer, slice, array-like or DataArray. Defaults to None.
            drop (bool, optional): Drop coordinates variables indexed by integers instead of making them scalar. Defaults to False.
        """
        var = self.get_value(x)
        result = var.data.isel(indexers=indexers, drop=drop)
        self.bind_value(y, PipelineVariable.from_result(result, var))

    # def where(self, y:ResultID, x:OperandID, cond:OperandID, /):
    # def cat(self, y:ResultID, x:OperandID, /, dim:str): coord np.ndarray data could be optional
    # def stack(self, y:ResultID, x:OperandID, /, coord:{name:str, data:np.ndarray}):

    #TODO: this should handle duplicate place names for higher admin levels. e.g. there are multiple Arlingtons (Arlington VA, Arlington TX, etc.)
    #      One possible way to handle is to take a tuple (place0, place1, place2, ...) (assert len(places) == admin_level + 1)
    @compile
    def reverse_geocode(self, y:ResultID, x:OperandID, /, places:list[str]=None, admin_level:int=0):
        """
        Adds a new 'admin<N>' (where N is the specified admin level) dimension to the data with slices for each specified place. 
        for each new place slice, masks out all data points not within the bounds of that place

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the data to split
            places (list[str], optional): the list of administrative boundaries to split the data by. Places may only be drawn from those at the specified admin level. If None, all places at the specified admin level will be used
            admin_level (int, optional): the admin level to split the data by. Defaults to 0 (i.e. country level).
        """
        # first make sure the data matches the specified resolution and frequency
        x = self.auto_regrid(x, allow_no_target=True)
        
        # get the raw data and lat/lon coordinates
        var = self.get_value(x)
        data = var.data
        lat: np.ndarray = data.lat.values
        lon: np.ndarray = data.lon.values

        # get the shapes for the specified admin level
        assert admin_level in range(0, 4), f'Invalid admin level: {admin_level}. Must be in range [0, 3]'
        if admin_level == 3:
            shapes = self.admin3
        else:
            shapes = self.admin2
        
        # if no places are specified, use all of them
        all_places = shapes[f'NAME_{admin_level}'].unique().tolist()
        if places is None:
            places = all_places
        else:
            # verify that all the places are valid
            all_places = set(all_places)
            for place in places:
                assert place in all_places, f'Invalid place: {place}. Must be one of: {all_places}'

        # set up new np array of shape [len(places), *data.shape]
        out_data = np.zeros((len(places), *data.shape), dtype=data.dtype)
        
        # needed for geometry_mask. cache here since we don't need to recompute it for each place
        transform = from_bounds(lon.min(), lat.min(), lon.max(), lat.max(), len(lon), len(lat))

        # keep track of the min/max lat/lon indices that are masked out
        min_lon_idx: int|None = None
        max_lon_idx: int|None = None
        min_lat_idx: int|None = None
        max_lat_idx: int|None = None

        for i, place in enumerate(places):
            self.print(f'processing {place}...')
            place_shapes = shapes[shapes[f'NAME_{admin_level}'] == place]
            
            # Generate a mask for the current place and flip it if necessary
            mask = geometry_mask(place_shapes['geometry'], transform=transform, invert=True, out_shape=(len(lat), len(lon)))
            if lat[0] < lat[-1]: mask = mask[::-1]
            if lon[0] > lon[-1]: mask = mask[:, ::-1]

            # apply the mask to the data
            masked_data = data.where(xr.DataArray(mask, coords={'lat':lat, 'lon':lon}, dims=['lat', 'lon']))

            # insert the masked data into the output array
            out_data[i] = masked_data

            # update the min/max lat/lon indices
            candidate_min_lon_idx = np.where(mask.any(axis=0))[0].min()
            min_lon_idx = candidate_min_lon_idx if min_lat_idx is None else min(min_lon_idx, candidate_min_lon_idx)
            candidate_max_lon_idx = np.where(mask.any(axis=0))[0].max()
            max_lon_idx = candidate_max_lon_idx if max_lon_idx is None else max(max_lon_idx, candidate_max_lon_idx)
            candidate_min_lat_idx = np.where(mask.any(axis=1))[0].min()
            min_lat_idx = candidate_min_lat_idx if min_lat_idx is None else min(min_lat_idx, candidate_min_lat_idx)
            candidate_max_lat_idx = np.where(mask.any(axis=1))[0].max()
            max_lat_idx = candidate_max_lat_idx if max_lat_idx is None else max(max_lat_idx, candidate_max_lat_idx)


        # combine all the data into a new DataArray with a country dimension
        out_data = xr.DataArray(
            out_data,
            dims=(f'admin{admin_level}', *data.dims),
            coords={
                f'admin{admin_level}': places,
                **data.coords,
            }
        )

        # crop the data to the minimum bounding box that contains all the places
        out_data = out_data.isel(lat=slice(min_lat_idx, max_lat_idx+1), lon=slice(min_lon_idx, max_lon_idx+1))
        
        # save the result to the pipeline namespace
        self.bind_value(y, PipelineVariable.from_result(out_data, var))

    @compile
    def mask_to_distance_field(self, y:ResultID, x:OperandID, /, include_initial_points:bool):
        """
        Convert a mask to a distance field. Distances are in kilometers

        Args:
            y (str): the identifier to bind the result to in the pipeline namespace
            x (str): the identifier of the mask to convert. Non-zero values are considered True, and zeros and NaNs are considered False
            include_initial_points (bool, optional): whether to include the initial points in the distance field. 
                If true, initial points will get a value of 0, otherwise they will be set to NaN. Defaults to True.
        """
        var = self.get_value(x)
        mask = var.data
        distances = mask_to_sdf(mask, include_initial_points=include_initial_points)

        self.bind_value(y, PipelineVariable.from_result(distances, var))
            


    @compile
    def save(self, x:OperandID, /, path:str):
        """
        Save the data to the given path

        Args:
            x (str): the identifier of the data to save
            path (str): the path to save the data to
        """
        var = self.get_value(x)
        var.data.to_netcdf(path)


    def execute(self):
        """Execute the pipeline"""
        for func, args, kwargs in self.steps:
            self.print(self.step_str((func, args, kwargs)))
            func(*args, **kwargs)

    @property
    def admin2(self):
        return get_admin2_shapes() # already cached

    @property
    def admin3(self):
        return get_admin3_shapes() # already cached

    # @property
    # def sf(self):
    #     """Get the country shapefile"""
    #     if self._sf is None:
    #         self.print(f'Loading country shapefile...')
    #         self._sf = gpd.read_file(f'{dirname(abspath(__file__))}/gadm_0/gadm36_0.shp')
    #     return self._sf
    
    def print(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)



from __future__ import annotations

from typing import Callable, TypeVar, Any, Iterable
from typing_extensions import ParamSpec

import inspect
import ast
import textwrap

import pdb

_P = ParamSpec('_P')
_R_co = TypeVar('_R_co', covariant=True)

def method_uses_prop(cls:type, method:Callable[_P, _R_co], prop:str, key:Callable[[Any], Callable[_P, _R_co]]=lambda m:m) -> bool:
    """
    Check if a method access a given property of the class in its source code.
    So far this is mainly used to check if a pipeline method requires GADM data which might not be downloaded yet.

    Args:
    - cls (type): The class that the method is attached to
    - method (callable): The method to check
    - prop (str): The name of the property to check for
    - key (callable): used for accessing the method directly from the `cls.__dict__`. usage is `key(cls.__dict__[method.__name__])`. E.g. pipeline methods are wrapped, so key can be used to unwrap them. Defaults to the identity function.
    """

    #check if the method is static. Static methods don't use properties
    class_attached = key(cls.__dict__[method.__name__])
    if isinstance(class_attached, staticmethod):
        return False

    # Fetch the source code of the method
    source = inspect.getsource(method)

    # get the parameter that represents self
    sig = inspect.signature(method)
    params = list(sig.parameters.values())
    assert len(params) > 0, f'Expected at least one parameter in method signature. Got: {sig}'
    self_param = params[0]
    
    # Dedent the source code to remove leading whitespaces
    dedented_source = textwrap.dedent(source)
    
    # Parse the dedented source code to get its AST
    tree = ast.parse(dedented_source)
    
    # Traverse the AST to look for instances where property 'a' of 'self' is accessed
    for node in ast.walk(tree):
        if isinstance(node, ast.Attribute) and node.attr == prop and isinstance(node.value, ast.Name) and node.value.id == self_param.name:
            return True
    return False


T = TypeVar('T')
def partition(elements:Iterable[T], condition:Callable[[T], bool]) -> tuple[list[T], list[T]]:
    """
    Partition a list of elements into two lists based on a condition.

    Args:
    - elements (iterable): The list of elements to partition
    - condition (callable): The condition to partition the elements on

    Returns:
    - tuple: A tuple containing two lists. The first list contains elements that satisfy the condition, the second list contains elements that don't satisfy the condition.
    """
    true_list = []
    false_list = []
    for element in elements:
        if condition(element):
            true_list.append(element)
        else:
            false_list.append(element)
    return (true_list, false_list)

import xarray as xr
import numpy as np
from scipy import stats
def nanmode(data:np.ndarray, dims:int|list[int]|None=None) -> np.ndarray:
    """
    Take the mode of the data along the specified dimensions

    Args:
    - data (np.ndarray): The data to take the mode of
    - dims (int|list[int], optional): The dimension(s) to take the mode over. If None, all dimensions are used. Defaults to None.

    Returns:
    - np.ndarray: The mode of the data along the specified dimensions
    """

    # ensure dims is a list
    if dims is None:
        dims = list(range(len(data.shape)))
    elif isinstance(dims, int):
        dims = [dims]

    # convert any negative dimensions to positive
    dims = [dim if dim >= 0 else len(data.shape) + dim for dim in dims]

    #transpose so that the dimensions to take the mode over are the last dimensions
    # reduce_dims, keep_dims = partition(var.data.dims, lambda dim: dim in dims)
    keep_dims = [i for i in range(len(data.shape)) if i not in dims]
    reduce_dims = dims
    data = np.transpose(data, keep_dims+reduce_dims)
    data = data.reshape(*(data.shape[:-len(reduce_dims)]+ (-1,)))

    # keep track of where all values in the mode slice are NaN
    # This is necessary because scipy 1.10 stats.mode would set a slice of all NaNs to 0
    # TODO: want to upgrade to scipy >= 1.11 since it properly handles all NaN slices 
    nan_mask = np.all(np.isnan(data), axis=-1)

    # take the mode
    data = np.ascontiguousarray(data)
    data = stats.mode(data, axis=-1, nan_policy='omit', keepdims=False).mode

    # data might not be an array
    # e.g. if the mode is a scalar, 
    # or mode had NaNs which returns some weird Masked Array type
    # TODO: TBD if this is the same in scipy >= 1.11
    data = np.array(data)

    # set all NaN slices back to NaN
    data[nan_mask] = np.nan

    return data

def xarray_mode(data:xr.DataArray, dims:str|list[str]|None=None) -> xr.DataArray:
    """
    Take the mode of the xarray DataArray along the specified dimensions

    Args:
    - data (xr.DataArray): The data to take the mode of
    - dims (list[str], optional): The dimensions to take the mode over. If None, all dimensions are used. Defaults to None.
    """
    # ensure dims is a list
    if dims is None:
        dims = [*data.dims]
    elif isinstance(dims, str):
        dims = [dims]

    # get the indices of the dimensions to reduce and the dimensions to keep
    reduce_dims, keep_dims = partition(data.dims, lambda dim: dim in dims)
    reduce_dims = [data.dims.index(dim) for dim in reduce_dims]
    keep_dims = [data.dims.index(dim) for dim in keep_dims]

    # take the mode
    result = nanmode(data.data, dims=reduce_dims)

    # convert result to xarray DataArray
    result = xr.DataArray(result, coords=[data.coords[i] for i in keep_dims], dims=[data.dims[i] for i in keep_dims])

    return result


from __future__ import annotations

from enum import Enum, auto
import numpy as np
import xarray as xr
from warnings import warn
from .spacetime import datetimeNoLeap_to_epoch
from .utilities import nanmode



class BinOffset(Enum):
    left = auto()
    # center = auto()  #TODO: for now don't use this
    # right = auto()   #TODO: for now don't use this

class RegridType(Enum):
    conserve = auto()
    min = auto()
    max = auto()
    mean = auto()    #TODO: consider removing in favor of just interp_or_mean
    median = auto()
    mode = auto()    #TODO: consider removing in favor of just nearest_or_mode
    interp_or_mean = auto()
    nearest_or_mode = auto()

# which methods weight the values of the bins on reduction
float_weighted_reduction_methods: set[RegridType] = {
    RegridType.mean,
    RegridType.interp_or_mean,
    RegridType.conserve,
}

# which methods take all values of bind unmodified on reduction 
boolean_reduction_methods: set[RegridType] = {
    RegridType.min,
    RegridType.max,
    RegridType.median,
    RegridType.mode,
    RegridType.nearest_or_mode,
}

assert all(method in float_weighted_reduction_methods or method in boolean_reduction_methods for method in RegridType), 'Some regrid methods are not accounted for in the float_weighted_reduction_methods or boolean_reduction_methods sets'


def compute_overlap(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """
    Given two arrays of bin edges, compute the amount of overlap between each bin in a and each bin in b

    Parameters:
        a (np.ndarray): The first array of bin edges
        b (np.ndarray): The second array of bin edges

    Returns:
        np.ndarray: A matrix of shape (len(a), len(b)) where each element is the proportion overlap between the corresponding bins in a and b
    """
    # Guarantee that a and b are in ascending order
    if (a_reversed := a[1] < a[0]):
        a = a[::-1]
    if (b_reversed := b[1] < b[0]):
        b = b[::-1]

    # Calculate overlap using ascending logic since they're both in the same order now
    a = a[:, None]
    overlap = np.maximum(0, np.minimum(a[1:], b[1:]) - np.maximum(a[:-1], b[:-1]))
    
    # Reverse the overlap matrix back so indices match the original arrays
    if a_reversed:
        overlap = overlap[::-1]
    if b_reversed:
        overlap = overlap[:, ::-1]
    
    return np.ascontiguousarray(overlap, dtype=np.float64)


def is_resolution_increase(overlaps:np.ndarray, inclusive:bool=False) -> bool:
    """
    Determine whether the resolution is increasing or decreasing based on the overlaps matrix
    """
    nonzero_indices = np.nonzero(overlaps)
    old_size = nonzero_indices[0].max() - nonzero_indices[0].min()
    new_size = nonzero_indices[1].max() - nonzero_indices[1].min()

    if inclusive:
        return new_size >= old_size
    else:
        return new_size > old_size

def get_interp_or_mean_overlaps(overlaps:np.ndarray, old_coords:np.ndarray, new_coords:np.ndarray) -> np.ndarray:
    """
    Convert the overlaps matrix to one that will perform:
    - interpolation for resolution increases
    - mean aggregation for resolution decreases
    """

    # resolution decrease just uses existing overlaps matrix
    if not is_resolution_increase(overlaps):
        return overlaps

    # compute the distances from each old cell to each new cell
    distances = np.abs(old_coords[:, None] - new_coords[None, :])

    #zero out everything except for the closest two cells in each column
    distance_ranks = np.argsort(distances, axis=0)
    distances[distance_ranks[2:], np.arange(distances.shape[1])] = np.nan

    #invert the distances so the closest cell has the largest weight
    distances = np.nansum(distances, axis=0)[None] - distances

    # normalize so that each column sums up to 1
    distances_sum = np.nansum(distances, axis=0)
    distances[:, distances_sum > 0] /= distances_sum[distances_sum > 0]

    # replace nans with 0
    distances[np.isnan(distances)] = 0

    return distances

def get_nearest_or_mode_overlaps(overlaps:np.ndarray, old_coords:np.ndarray, new_coords:np.ndarray) -> np.ndarray:
    """
    Get an overlaps matrix that just selects the item nearest to the location of the new bin
    - nearest neighbor for resolution increases
    - mode aggregation for resolution decreases
    """
    
    # resolution decrease just uses existing overlaps matrix
    if not is_resolution_increase(overlaps):
        return overlaps

    # compute the distances from each old cell to each new cell
    distances = np.abs(old_coords[:, None] - new_coords[None, :])

    #zero out everything except for the closest cell in each column, and set the closest cell to 1
    distance_ranks = np.argsort(distances, axis=0)
    distances[distance_ranks[1:], np.arange(distances.shape[1])] = 0
    distances[distances > 0] = 1

    return distances


def get_bins(coords:np.ndarray, offset:BinOffset) -> np.ndarray:
    """
    Given an array of evenly spaced coordinates, return the bin edges

    Parameters:
        coords (np.ndarray): The array of evenly spaced coordinates
        offset (BinOffset): Whether the bins should be left, centered, or right of the coordinates

    Returns:
        np.ndarray, np.ndarray: The bin edges, and the deltas between each bin
    """
    if len(coords) < 2:
        raise ValueError('Cannot infer bin edges from a single coordinate')
        
    deltas = coords[1:] - coords[:-1]
    if not np.allclose(deltas, deltas[0]):
        #This happens for e.g. monthly timestamps where months can have different numbers of days
        warn(f'coords are not evenly spaced. Got spacings: {np.unique(deltas)}', RuntimeWarning)

    if offset == BinOffset.left:
        bins = np.concatenate([[coords[0] - deltas[0]], coords])
    elif offset == BinOffset.center:
        bins = np.concatenate([[coords[0] - deltas[0]], coords]) + np.concatenate([[deltas[0]], deltas, [deltas[-1]]]) / 2
    elif offset == BinOffset.right:
        bins = np.concatenate([coords, [coords[-1] + deltas[-1]]])

    # update the deltas to include the edges
    deltas = np.diff(bins)

    return bins, deltas



def regrid_1d(
        data:xr.DataArray,
        new_coords:np.ndarray,
        dim:str,
        offset:BinOffset=BinOffset.left,
        aggregation=RegridType.interp_or_mean,
        wrap:tuple[float,float]=None, #TBD format for this...
        low_memory:bool=False,
    ) -> xr.DataArray:
    
    # grab the old coords and data (copy data so we don't modify the original)
    old_coords = data[dim].data
    old_data = data.data.copy()

    # convert time coords to epoch timestamps
    if dim == 'time':
        old_coords = np.array([datetimeNoLeap_to_epoch(time) for time in old_coords])
        new_coords_copy = new_coords.copy()
        new_coords = np.array([datetimeNoLeap_to_epoch(time) for time in new_coords])

    # get the bin boundaries for the old and new data
    old_bins, old_deltas = get_bins(old_coords, offset)
    new_bins, _ = get_bins(new_coords, offset)
    
    #compute the amount of overlap between each old bin and each new bin
    overlaps = compute_overlap(old_bins, new_bins)
    
    # normalization so that overlaps measure the percentage of each old cell that overlaps with the new bins
    overlaps /= np.abs(old_deltas[:, None])

    # handle aggregation methods that use a modified overlaps matrix
    if aggregation == RegridType.interp_or_mean:
        overlaps = get_interp_or_mean_overlaps(overlaps, old_coords, new_coords)
    elif aggregation == RegridType.nearest_or_mode:
        overlaps = get_nearest_or_mode_overlaps(overlaps, old_coords, new_coords)

    # ensure the dimension being operated on is the last one
    original_dim_idx = data.dims.index(dim)
    if original_dim_idx != len(data.dims) - 1:
        old_data = np.moveaxis(old_data, original_dim_idx, -1)

    # crop out of bounds data from old data/overlaps
    unused_rows = np.all(overlaps == 0, axis=1)
    old_data = old_data[..., ~unused_rows]
    overlaps = overlaps[~unused_rows]

    #ensure the data is contiguous in memory
    old_data = np.ascontiguousarray(old_data)
    overlaps = np.ascontiguousarray(overlaps)

    #set any data that is nan to 0, and keep track of where the nans were
    validmask = ~np.isnan(old_data)
    old_data[~validmask] = 0

    # hacky way to deal with nans not propagating correctly under mode aggregation
    # TODO: is this correct for nearest interpolation?
    if aggregation == RegridType.mode or aggregation == RegridType.nearest_or_mode:
        if not np.all(validmask):
            old_data[~validmask] = float('-inf') #TODO: this line will fail on integer data. but validmask should be all True in that case, so probably no problem...

    #perform the regridding on the data, and replace any nans
    result = regrid_1d_reducer(old_data, overlaps, aggregation, low_memory)
    replace_nans(result, validmask, overlaps)

    # move the dimension back to its original position
    result = np.moveaxis(result, -1, original_dim_idx)
    result = np.ascontiguousarray(result)

    #convert time coords back to cftime.DatetimeNoLeap
    if dim == 'time':
        new_coords = new_coords_copy

    #convert back to xarray, with the new coords
    result = xr.DataArray(result, coords={**data.coords, dim:new_coords}, dims=data.dims)

    return result





def regrid_1d_reducer(old_data:np.ndarray, overlaps:np.ndarray, aggregation:RegridType, low_memory:bool=False) -> np.ndarray:
    """
    Perform the actual regridding reduction over the output bins, according to the aggregation method
    """
    
    # low memory mode uses less memory, but is less accurate
    if low_memory:
        old_data = old_data.astype(np.float32)
        overlaps = overlaps.astype(np.float32)

    overlap_mask = overlaps > 0
    cols = np.any(overlap_mask, axis=0)
    starts = overlap_mask.argmax(axis=0)
    ends = overlap_mask.shape[0] - (overlap_mask[::-1]).argmax(axis=0)

    #determine the length of the longest column
    max_col_length = np.max(ends[cols] - starts[cols])

    #set the starts and ends of the empty columns to 0:max_col_length
    starts[~cols] = 0
    ends[~cols] = max_col_length

    #extend all columns so that they are of length max_col_length
    ends[cols] = (starts[cols] + max_col_length).clip(max=overlap_mask.shape[0])
    starts[cols] = (ends[cols] - max_col_length).clip(min=0)

    #convert overlaps/overlap_mask so that unselected locations are nan
    overlaps[~overlap_mask] = np.nan
    overlap_mask[~overlap_mask] = np.nan
    # overlap_mask = overlap_mask.astype(np.float32)
    # overlap_mask[overlap_mask==0] = np.nan #TODO: this is a bool array, so np.nan here will be set to true...
    # set up selectors for the indices of each bin to aggregate over
    col_selector = np.arange(max_col_length) + starts[:, None]
    row_selector = np.arange(overlap_mask.shape[1])[:, None]

    #construct a matrix holding all the input values for each bin in the output
    unmasked_binned_data = old_data[..., col_selector]
    
    if aggregation in float_weighted_reduction_methods:
        # weight values in the bin before aggregating
        bin_mask = overlaps[col_selector, row_selector]
    else:
        # take all values in the bin unmodified for aggregation
        bin_mask = overlap_mask[col_selector, row_selector]
    
    # mask (or weight) the sets of values in the bins
    binned_data = unmasked_binned_data * bin_mask

    # perform reduction over bins according to aggregation method
    if aggregation == RegridType.min:
        result = np.nanmin(binned_data, axis=-1)
    elif aggregation == RegridType.max:
        result = np.nanmax(binned_data, axis=-1)
    elif aggregation == RegridType.mean or aggregation == RegridType.interp_or_mean:
        result = np.nansum(binned_data, axis=-1) / np.nansum(bin_mask, axis=-1)
    elif aggregation == RegridType.median:
        result = np.nanmedian(binned_data, axis=-1)
    elif aggregation == RegridType.mode:
        result = nanmode(binned_data, dims=-1)
    elif aggregation == RegridType.nearest_or_mode:
        if binned_data.shape[-1] == 1:  # select the only value in each bin
            result = binned_data[..., 0]
        else:
            result = nanmode(binned_data, dims=-1)
    elif aggregation == RegridType.conserve:
        result = np.nansum(binned_data, axis=-1)
    else:
        raise NotImplementedError(f'Unrecognized regrid aggregation method: {aggregation}. Expected one of: {[*RegridType.__members__.values()]}')

    return result
    

def replace_nans(result:np.ndarray, validmask:np.ndarray, overlaps:np.ndarray):
    if np.all(validmask):
        return

    #TODO: handling MemoryError when running out of memory
    nan_accumulation = validmask.astype(np.float32) @ (overlaps > 0).astype(np.float32)
    result[nan_accumulation == 0] = np.nan
